#include <stdio.h>
#include <string.h>
#include <argtable2.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>
#include <linux/tcp.h>        // TCP_MAXSEG etc.
#include <linux/net.h>        // ioctl - socket enums
#include <netinet/in.h>       // IPPROTO def's
#include <sys/ioctl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
static char *numbers= NULL;


// Global pointer to arg-table - needed for sighandler
typedef int const bool;
typedef void * table_vptrs[];

table_vptrs *parg_table;

static int   argtablesize;

static int is_step;
static int is_quiet;

void hit()
{   
   if (is_step)
   {
      printf(".");   
      getchar();
   }
}


void sig_handler(int sig)
{
   printf("Stopped with CTRL-C - cleaning up (signal %d)\n", sig);
   arg_freetable(*parg_table, argtablesize);
   exit(0);
}

void cleanup()
{
   if (!is_quiet)
      printf("Cleaning up\n");

   hit();
   arg_freetable(*parg_table, argtablesize);
   exit(0);
}

int do_connect(int socket, const char *ip, const bool port)
{
   if (!is_quiet)
      printf("Connecting to ip: %s, port: %d on socket: %d\n", ip, port, socket);

   hit();

   struct sockaddr_in6 dest_addr;         // will hold the destination addr
   dest_addr.sin6_family = AF_INET6;      // host byte order 
   dest_addr.sin6_port = htons(port);      // short, network byte order 
   
   int err = inet_pton(AF_INET6, ip, &dest_addr.sin6_addr);
   if (err != 1)
   {
      printf("Err: %d Illegal address.\n", err);
      exit(err);
   }

   int retval;
   retval = connect(socket, (struct sockaddr *)&dest_addr, sizeof(dest_addr));   
      
   if (retval)
   {
      printf("Connecting to ip: %s, port %d on socket %d gave error: %d\n", ip, port, socket, retval);
      cleanup();
   }  
   return retval;
}

int do_bind(int socket, const char *ip, int port)
{
   if (!is_quiet)
      printf("Binding to ip: %s, port: %d on socket: %d\n", (ip) ? ip : "localhost", port, socket);

   hit();

   struct sockaddr_in6 src_addr;          // will hold the source addr
   src_addr.sin6_family = AF_INET6;       // host byte order 
   src_addr.sin6_port = htons(port);       // short, network byte order 
   if (ip)
   {
      int err = inet_pton(AF_INET6, ip, &src_addr.sin6_addr);
      {
         printf("Illegal address.\n");
         exit(err);
      } 
   }
   else
      src_addr.sin6_addr = in6addr_any;

   int retval;
   retval = bind(socket, (struct sockaddr *)&src_addr, sizeof(src_addr));      
      
   if (retval)
   {
      printf("Binding to ip: %s, port %d on socket %d gave error: %d\n",  (ip) ? ip : "localhost", port, socket, retval);
      cleanup();
   }
      
   return retval;
}

int do_listen(int socket)
{
   if (!is_quiet)
      printf("Listen to socket %d\n", socket);

   hit();

   int retval = listen(socket, 5); // Backlog of 5 sockets
   
   if (retval)
   {
      printf("Listening to socket %d gave error: %d\n", socket, retval);
      cleanup();
   }
     
   return retval;
}

int do_socket()
{
   if (!is_quiet)
      printf("Will now open a socket\n");

   hit();
      
   int sock = socket(AF_INET6, SOCK_STREAM, 0);
   if (sock <= 0)
   {
      printf("Opening socket gave error: %d\n",sock);
      cleanup();
   }
     
   return sock;   
}

int do_close(int socket)
{
   if (!is_quiet)
      printf("Will now close socket %d\n", socket);

   hit();
      
   int retval = close(socket);
   if (retval)
   {
      printf("Closing socket %d gave error: %d\n", socket, retval);
      cleanup();
   }
     
   return retval;   
}

int do_accept(int socketLocal)
{
   if (!is_quiet)
      printf("Accept on socket %d\n", socketLocal);

   hit();

   struct sockaddr_in6 remote;
   socklen_t adr_len = sizeof(remote);
   
   int retval =  accept(socketLocal, (struct sockaddr *)&remote, &adr_len);      
      
   if (retval <= 0)
   {
      printf("Accepting on socket %d gave error: %d\n", socketLocal, retval);
      cleanup();
   }
   else if (!is_quiet)
   {
   	char ipstr[100];
   	inet_ntop(AF_INET6, (void*) &remote.sin6_addr, ipstr, sizeof(ipstr));
      	printf("Got accept on socket %d with: %s port %d - new socket %d\n", socketLocal, 
                ipstr, ntohs(remote.sin6_port), retval);
   }
   return retval; // positive value is new socket
}

int do_send(int socket, int bytes)
{  
   int err;
   
   if (!is_quiet)
      printf("Plan to send %d bytes on socket %d\n", bytes, socket);

   hit();
   
   if ((err = send(socket, numbers, bytes, 0)) < 0)
   {
      printf("Could not send data on socket: %d, %d bytes. Error: %d\n", socket, bytes, err);
      cleanup();
   }
   
   return 0;
}

int do_recv(int socket, int bytes)
{
   const long long progress_when = 1000000000;
   static long long total_count = 0;
   long progress1, progress2;
   char *pbuf = NULL;
   char nzbuf[bytes];
   int total = 0;
   
   if (!is_quiet)
      printf("Plan to receive %d bytes on socket %d\n", bytes, socket);

   hit();
 
   int received = recv(socket, nzbuf,bytes-total, 0);
   if (!is_quiet)   
      printf("Received %d bytes at address %p:\n", received, nzbuf);

   else if (received < 0)
   {
      if (errno == EWOULDBLOCK)
      {
         printf("recv is either non-blocking or timed out - sleep 1s and try again\n");
         sleep(1);
      }
      else
      {
         printf("Could not recv data on socket: %d, %d bytes sofar, errno: %d, pbuf=%p\n", 
                 socket, total, errno, pbuf);
         cleanup();
      }
   }
   else
   {
      progress1 = total_count / progress_when;
      total_count += received;
      progress2 = total_count / progress_when;
      if (progress1 != progress2)
         printf("Received %18lld bytes\n", total_count);
   }

   if (!is_quiet)         
   {
      char tmpb[bytes+1];
      int to_copy =  ((int) sizeof(tmpb) > received) ? received : (int) sizeof(tmpb)-1;
      //printf("bytes: %d, to_copy: %d, received: %d\n", bytes, to_copy, received);
      memcpy(tmpb, nzbuf, to_copy);         
      tmpb[to_copy] = '\0';
      printf("%s\n", tmpb);
   }

   return received;
}



void TestNames(int socketLocal)
{
   struct sockaddr_in6 sock_addr;
   socklen_t adr_len = sizeof(sock_addr);
   char ipstr[100];

   getpeername(socketLocal,(struct sockaddr *) &sock_addr, &adr_len);

   inet_ntop(AF_INET6, (void*) &sock_addr.sin6_addr, ipstr, sizeof(ipstr));

   printf("getpeername: IP= %s, port= %d, adr_len= %d \n", ipstr
         , ntohs(sock_addr.sin6_port), adr_len); 

}


// Build the static string used in send
void build_numbers(int bytes)
{ 
   numbers = (char*) malloc(bytes+10);
   char tmp[10];
   numbers[0] = '\0';
   
   int i;
   int charcount=0;
   
   for(i=0; charcount < bytes; i++)
   {
      charcount += sprintf(tmp,"%4x ",i);
      strcat(numbers, tmp);
   }
}

int mymain(const int loops,   const bool is_recv,  const bool is_send,  const bool is_client,
           const int port,    const char *ip,      const bool is_term,
           const int bytes)
{
   char me[80];
   
   gethostname(me, sizeof(me));
  
   if (is_send)
      build_numbers(bytes);
   
//   printf("Hello from %s\n", me);
//   
//   printf("You selected:\n");
//   printf("Loops: %d, recv: %d, send: %d, client: %d, fake: %d\n", loops, is_recv, is_send, is_client, is_fake);
//   printf("IP: %s, Port: %d, quiet: %d, term: %d, bytes: %d\n", ip, port, is_quiet, is_term, bytes);

   const bool is_forever = (loops == 0);
   int count;
   long long transmitted_total = 0;
   int transmitted;
   struct timeval start, end;
      
   // Install signal handler - enables termination without using a system call
   // every loop-cycle for checking keyboard.
   signal (SIGINT, sig_handler);
//   signal (SIGQUIT, sig_handler);
//   signal (SIGTERM, sig_handler);
//   signal (SIGABRT, sig_handler);
   
   int sock_comm;
   int sock1 = do_socket();      
   


   if (is_client)
   {
      do_connect(sock1, ip, port);
      sock_comm = sock1;
   }
   else // server
   {
      do_bind(sock1, 0, port);             
      do_listen(sock1); 
      sock_comm = do_accept(sock1);   
   }

   TestNames(sock_comm);

   gettimeofday(&start, NULL);

   for (count=0; is_forever || count < loops; count++)
   {
      if (!is_quiet) 
      printf("Loop %5d\n", count);
 
      if (is_send)
      {
         do_send(sock_comm, bytes);
	 transmitted_total += bytes;
      }
      if (is_recv)
      {            
         if ((transmitted = do_recv(sock_comm, bytes)) == 0)
	    break; // When receiving 0 bytes the other end has closed
         transmitted_total += transmitted;
      }
   }

 
      
   gettimeofday(&end, NULL);
   // Allow Tx to complete 
   //sleep(2); 
               
   double time_spent = difftime(end.tv_sec, start.tv_sec);      
   time_spent += ((end.tv_usec-start.tv_usec)/1000000.0);
   double rate = (8.0 * ((double)transmitted_total)/time_spent)/1000000.0;
   printf("Totally transmitted: %lld in %d loops in %f seconds =  %6.3g Mb/s\n", transmitted_total, count, time_spent, rate);
   printf("Time was measured from just after 'connect'/'accept' until now\n");

   if (is_term)
   {
      
      do_close(sock_comm);
      if (is_client)
         do_close(sock1);
   }
    
   return 0;
}

int main(int argc, char **argv)
{
   struct arg_lit    *recv    = arg_lit0("r","recv",                          "receive data");
   struct arg_lit    *send    = arg_lit0("s","send",                          "send data (time of day)");
   struct arg_lit    *quiet   = arg_lit0("q","quiet",                         "suppress messages to gain speed");
   struct arg_lit    *conn    = arg_lit0("c","client",                        "Connect as client - otherwise server(default)");    
   struct arg_lit    *term    = arg_lit0("t","terminate",                     "terminate (close) actively or passively (default)");       
   struct arg_lit    *step    = arg_lit0("o","One",                           "One socket command - then wait for return - etc");       
   struct arg_lit    *help    = arg_lit0(NULL,"help",                         "print this help and exit");
   struct arg_lit    *version = arg_lit0(NULL,"version",                      "print version information and exit");
   struct arg_int    *loop    = arg_int0("l","loop",NULL,                     "loop n times (0 = infinite - ctrl-C to stop)");
   struct arg_int    *bytes   = arg_int0("b","bytes",NULL,                    "Bytes of data in each sent message");
   struct arg_int    *port    = arg_int0("p","port",NULL,                     "port to connect to or listen to (default 12000)");
// struct arg_file   *infiles = arg_filen(NULL,NULL,NULL,1,argc+2,            "input file(s)");
   struct arg_str    *ip      = arg_str0("i","ip-address","x.y.z.v",          "ip to connect to (when client)");       
   struct arg_end    *end     = arg_end(20);


//   void* argtable[]  = {recv,send,quiet,help,version,loop,port,conn,term,ip,   end};
    table_vptrs argtable = {recv,send,bytes,quiet,help,version,loop,port,conn,term,ip,step,   end};
    
   const char* progname = argv[0];
   int nerrors;
   int exitcode=0;

   /* verify the argtable[] entries were allocated sucessfully */
   if (arg_nullcheck(argtable) != 0)
   {
      /* NULL entries were detected, some allocations must have failed */
      printf("%s: insufficient memory\n",progname);
      exitcode=1;
      goto exit;
   }

   /* set any command line default values prior to parsing */
   loop->ival[0]=1;
   port->ival[0]=12000;
   ip->sval[0]   = "::1";
    
//    outfile->filename[0]="-";

   /* Parse the command line as defined by argtable[] */
   nerrors = arg_parse(argc,argv,argtable);

   /* special case: '--help' takes precedence over error reporting */
   if (help->count > 0)
   {
      printf("Usage: %s\n", progname);
      arg_print_syntax(stdout,argtable,"\n");
      printf("This program is used for testing the IPB library\n");
      arg_print_glossary(stdout,argtable,"  %-35s %s\n");
      exitcode=0;
      goto exit;
   }

    /* special case: '--version' takes precedence error reporting */
   if (version->count > 0)
   {
      printf("'%s' Using the \"argtable\" command line argument parser.\n",progname);
      exitcode=0;
      goto exit;
   }

   /* If the parser returned any errors then display them and exit */
   if (nerrors > 0)
   {
      /* Display the error details contained in the arg_end struct.*/
      arg_print_errors(stdout,end,progname);
      printf("Try '%s --help' for more information.\n",progname);
      exitcode=1;
      goto exit;
   }

    /* special case: no command line options induces brief help */
   if (argc==1)
   {
      printf("Try '%s --help' for more information.\n",progname);
      exitcode=0;
      goto exit;
   }

   parg_table    = &argtable;
   argtablesize  = sizeof(argtable)/sizeof(argtable[0]);

   is_step  = step->count > 0;
   is_quiet = quiet->count > 0;


   /* normal case: take the command line options at face value */
   exitcode = mymain(loop->ival[0], recv->count,   send->count,   conn->count,
                     port->ival[0], ip->sval[0],   term->count,
                     bytes->ival[0]);

   exit:
   arg_freetable(*parg_table, argtablesize);

   return exitcode;
}

